#include <fstream>
#include <sstream>
#include <stdlib.h>
#include "Bit_stream.h"

using namespace std;

struct Argument_error {
    string description;
    Argument_error(string desc) : description(desc) {}
};

void parse_arguments(int argc, char * argv[], const char ** filename, long *offset, long *block_size) {
    if (argc < 2) {
        throw Argument_error("missing file name");
    }
    *filename = argv[1];
    *offset = 0;
    *block_size = 0x8000;

    for (int argno = 2; argno < argc; argno++) {
        if (argv[argno][0] == '-' && argv[argno][1] != '\0' &&
                argv[argno][2] == '\0') {
            switch (argv[argno][1]) {
                case 'o':
                    if (argno + 1 >= argc) {
                        throw Argument_error("-o needs an argument");
                    }

                    {
                        stringstream argstring(argv[++argno]);
                        argstring >> hex >> *offset;
                        /* TODO: check that conversion went ok */
                    }
                    continue;
                case 'b':
                    if (argno + 1 >= argc) {
                        throw Argument_error("-b needs an argument");
                    }

                    {
                        stringstream argstring(argv[++argno]);
                        argstring >> hex >> *block_size;
                        /* TODO: check that conversion went ok */
                    }
                    continue;
            }
        }

        throw Argument_error(
                string("don't know what to do with argument \"") + argv[argno] + "\"");
    }
}

void parse_xma(const char * file_name, long offset, long block_size);

int main(int argc, char * argv[]) {
    const char * file_name;
    long offset, block_size;

    cout << "Experimental XMA parser 0.0 (blame hcs)" << endl << endl << flush;
    try {
        parse_arguments(argc, argv, &file_name, &offset, &block_size);
    } catch (Argument_error ae) {
        cerr << ae.description << endl;
        cerr << "usage: xma_parse filename [-o offset] [-b block size]" << endl;
        exit(EXIT_FAILURE);
    }

    cout << "filename: " << file_name << endl;
    cout << "offset: " << hex << offset << dec << endl;
    cout << "block size: " << hex << block_size << dec << endl;

    parse_xma(file_name, offset, block_size);
}

enum { XMA_packet_size = 0x800, XMA_packet_header_size = 4 };

struct XMA_packet_header {
    Bit_uint<6> frame_count;
    Bit_uint<15> skip_bits;
    Bit_uint<3> metadata;
    Bit_uint<8> packet_skip;

    friend Bit_stream& operator>> (Bit_stream& bs, XMA_packet_header &ph) {
        return bs >> ph.frame_count >> ph.skip_bits >>
                     ph.metadata >> ph.packet_skip;
    }
};

void parse_xma(const char * file_name, long offset, long block_size) {
    ifstream is(file_name, ios::binary);

    long last_offset = offset + block_size;
    for (int packet_number = 0; offset < last_offset; packet_number++) {
        XMA_packet_header ph;

        {
            Bit_stream packet_header_stream(is);

            is.seekg(offset);

            packet_header_stream >> ph;
        }

        cout << "Packet #" << packet_number << endl;
        cout << "Frame Count     " << ph.frame_count << endl;
        cout << "Skip Bits       " << ph.skip_bits << endl;
        cout << "Metadata        " << ph.metadata << endl;
        cout << "Packet Skip     " << ph.packet_skip << endl;

        Bit_stream frame_header_stream(is,
                // consecutive
                (XMA_packet_size - XMA_packet_header_size) * 8,
                // skip
                (XMA_packet_header_size + ph.packet_skip * XMA_packet_size)
                * 8
                );

        // skip initial bits (overflow from a previous packet)
        for (unsigned int i = 0; i < ph.skip_bits; i++) frame_header_stream.get_bit();

        for (unsigned int frame_number = 0; frame_number < ph.frame_count; frame_number++) {
            Bit_uint<15> frame_bits;
            frame_header_stream >> frame_bits;

            cout << "   Frame #" << frame_number << endl;
            cout << "   Size " << frame_bits << endl;

            // read frame content (throw away for now)
            unsigned int i;
            cout << hex;
            for (i = 0; i + 4 <= frame_bits - 15; i += 4) {
                Bit_uint<4> nybble;
                frame_header_stream >> nybble;
                cout << nybble;
            }
            cout << " ";
            for (; i < frame_bits - 15; i++) {
                cout << (frame_header_stream.get_bit() ? '1' : '0');
            }
            cout << dec << endl;
        }

        cout << endl;
        offset += (ph.packet_skip + 1) * XMA_packet_size;
    }
}
